package ir.instr;

import ir.type.Type;
import ir.value.BasicBlock;
import ir.Value;

import java.util.HashMap;
import java.util.LinkedList;

public class Phi extends Instr {
    public HashMap<BasicBlock, Value> values;   //指令块和value的键值对
    private final Type phiType;

    private Alloca owner;

    public Alloca getOwner() {
        return owner;
    }

    public void setOwner(Alloca owner) {
        this.owner = owner;
    }

    public HashMap<BasicBlock, Value> getValues() {
        return values;
    }

    public void setValues(HashMap<BasicBlock, Value> values) {
        this.values = values;
    }

    public Phi(Type phiType, BasicBlock parent) {
        super(new LinkedList<>(), phiType, "tmpx", parent);
        this.phiType = phiType;
        this.values = new HashMap<>();
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(getFullName()).append(" = ");
        sb.append("phi ").append(phiType).append(" ");
        for (BasicBlock basicBlock : values.keySet()) {
            sb.append("[ ");
            sb.append(values.get(basicBlock).getFullName());
            sb.append(", ");
            sb.append("%" + basicBlock.getName());
            sb.append(" ], ");
        }
        sb.delete(sb.length() - 2, sb.length());
        return sb.toString();
    }


    public void addBlock2Value(BasicBlock block, Value value) {
        values.put(block, value);
        this.addValue(value);
        value.addUser(this);
    }

    public void removeRedundant() {
        //TODO: 2023/7/20
    }

    @Override
    public Instr clone(BasicBlock basicBlock, HashMap<Value, Value> valueHashMap) {
        //TODO:如有需要可以补写
        return null;
    }

    @Override
    public void reClone(HashMap<Value, Value> valueHashMap) {
        for (BasicBlock basicBlock : values.keySet()) {
            if (valueHashMap.containsKey(basicBlock)) {
                values.put((BasicBlock) valueHashMap.get(basicBlock),
                    valueHashMap.get(values.get(basicBlock)));
            } else if (valueHashMap.containsKey(values.get(basicBlock))) {
                values.put((BasicBlock) valueHashMap.get(basicBlock),
                    valueHashMap.get(values.get(basicBlock)));
            }
        }
    }

    public void checkAndRemoveBlock(BasicBlock basicBlock) {
        if (values.containsKey(basicBlock)) {
            this.getOperands().remove(values.get(basicBlock));
            values.get(basicBlock).users.remove(this);
            values.remove(basicBlock);
        }
    }

    public void replaceBlockWith(BasicBlock ori, BasicBlock newblock) {
        if(values.containsKey(ori)){
            Value tmp = values.get(ori);
            values.remove(ori);
            values.put(newblock, tmp);
        }
    }

    @Override
    public void replaceOp(Value old, Value newop) {
        super.replaceOp(old, newop);
        for(var bb : values.keySet()){
            if(values.get(bb) == old){
                values.put(bb, newop);
            }
        }
    }
}
