package ir.pass.analyze;

import ir.value.BasicBlock;
import ir.value.Module;
import tools.BlockAnalyzer;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Stack;

public class LoopAnalyzer {
    private static ArrayList<Loop> currentLoops;
    private static LinkedHashMap<BasicBlock, Boolean> visited = new LinkedHashMap<>();

    //必须先运行domAnalyzer，获得支配树
    public static void analyzeLoop(Module module){
        for(var func : module.functions.values()){
            func.loops.clear();
        }
        for(var func : module.functions.values()){
            for(var block : func.blocks){
                block.setLoopDepth(0);
                block.loop = null;
            }
        }

        for(var func : module.functions.values()){
            currentLoops = func.loops;
            if(func.blocks.size()==0){
                continue;
            }

            visit(func.blocks.get(0));
            for(Loop loop : currentLoops){
                loop.updateLevel();
            }
        }
    }

    private static void visit(BasicBlock block){
        for( var subb : block.getIdominate()){
            visit(subb);
        }
        Stack<BasicBlock> keyEdges = new Stack<>();
        for(var pred : block.getPreds()){
            if(block.getDominate().contains(pred)){
                keyEdges.add(pred);
            }
        }
        if(keyEdges.size()==0){
            return;
        }
        Loop loop = new Loop();
        currentLoops.add(loop);
        visited = new LinkedHashMap<>();
        visited.put(block, true);
        block.loop = loop;
        loop.blocks.add(block);
        while(keyEdges.size()>0){
            var tmp = keyEdges.pop();
            if(visited.containsKey(tmp))continue;
            visited.put(tmp, true);
            if(tmp.loop == null){
                for(var pre : tmp.getPreds()){
                    if (visited.containsKey(pre)) {
                        continue;
                    }
                    keyEdges.add(pre);
                }
                tmp.loop = loop;
                loop.blocks.add(tmp);
            }else{
                for(var pre : tmp.getPreds()){
                    if (visited.containsKey(pre) || tmp.loop.blocks.contains(pre)) {
                        continue;
                    }
                    keyEdges.add(pre);
                }
                tmp.loop.father = loop;
                loop.children.add(tmp.loop);
            }
        }
    }
}
