package ir.pass.analyze;

import java.util.HashMap;
import java.util.LinkedHashMap;

import ir.Value;
import ir.value.*;
import ir.value.user.Function;
import ir.value.Module;
import ir.instr.*;

public class SideEffectAnalyzer
{
    /**
     * 这个表示需要被处理的 func
     */
    private static final LinkedHashMap<Function, Boolean> processed = new LinkedHashMap<>();
    private static final LinkedHashMap<Function, Boolean> visited = new LinkedHashMap<>();

    public static void calSideEffect(Module module)
    {
//        System.out.println("TEST!!!!!!!!!!!!!!!");
        initCallGraph(module);
        Function main = module.getFunction("main");
        process(main);
//        for(var function:module.functions.values()){
//            System.out.println(function.getFullName());
//            System.out.println(function.hasSideEffect);
//        }
        main.hasSideEffect=true;
    }

    private static void initCallGraph(Module module)

    {
        for (Function function : module.functions.values())
        {
            function.getCallees().clear();
            function.hasSideEffect = false;
            visited.put(function, false);
            // 如果一个函数是运行时库函数，那么就是有副作用的
            if (function.blocks.size()==0)
            {
                function.hasSideEffect = true;
                processed.put(function, true);
                continue;
            }

            // processed 可能是另一种形式的 visited
            processed.put(function, false);

            for (BasicBlock block : function.blocks)
            {
                for (Value instr : block.getInsts())
                {
                    // 遍历 call 指令
                    if (instr instanceof Call)
                    {
                        Function calledFunction = ((Call) instr).getFunction();
                        // 如果不是运行时库函数，加入callee
                        if (calledFunction.blocks.size()!=0)
                        {
                            function.getCallees().add(calledFunction);
                        }
                        // 认为调用运行时库函数是有影响的
                        else
                        {
                            function.hasSideEffect = true;
                            processed.put(function, true);
                        }
                    }

                    // 如果是 store 那么就是有影响的
                    if (instr instanceof Store)
                    {
                        function.hasSideEffect = true;
                        processed.put(function, true);
                    }
                }
            }
        }
    }

    private static boolean process(Function function)
    {
        boolean hasSideEffect = false;
        visited.put(function, true);
        // 如果这个函数已经被处理了
        if (processed.get(function))
        {
            hasSideEffect = function.hasSideEffect;
            // 遍历它的调用者
            for (Function callee : function.getCallees())
            {
                if (!processed.get(callee) && !visited.get(callee))
                {
                    process(callee);
                }
            }
        }
        // 还没有被处理
        else
        {
//            System.out.println(function.getFullName());
            for (Function callee : function.getCallees())
            {
//                System.out.println("Child");
//                System.out.println(callee.getFullName());
                if (callee.hasSideEffect ||
                        (!processed.get(callee) && !visited.get(callee) && process(callee)))
                {
//                    System.out.println(callee.hasSideEffect);
//                    System.out.println(callee.getFullName());
                    hasSideEffect = true;
                }
            }
        }

        function.hasSideEffect=hasSideEffect;
        processed.put(function, true);
        return hasSideEffect;
    }
}
