package ir.pass.opt;
import ir.Value;
import ir.value.Module;
import ir.value.*;
import ir.instr.*;
import ir.value.user.Function;
import ir.pass.analyze.*;
import tools.ModuleFlattener;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;

/*
    store call br
*/
public class DeadCodeEmit implements Pass{
    LinkedHashSet<Function> usefulFunctionClosure = new LinkedHashSet<>();
    LinkedHashSet<Function>  uselessFunction = new LinkedHashSet<>();
    LinkedHashSet<Value> usefulInstrClosure = new LinkedHashSet<>();
    LinkedList<User> queue = new LinkedList<>();
    @Override
    public void run(Module module){
      SideEffectAnalyzer.calSideEffect(module);
      deadCodeEmitInstrLevel(module);
      deadCodeEmitFunctionLevel(module);
      deadCodeEmitVariableLevel(module);
      deadCodeEmitParamLevel(module);
//      ModuleFlattener.flattenModule(module);
    }

    public void deadCodeEmitParamLevel(Module module){
        //TODO DeadParamEmit 2023/7/26
    }
    public void deadCodeEmitVariableLevel(Module module){
        module.init.getInsts().removeIf(variable -> variable.getUsers().isEmpty());
    }
    public void deadCodeEmitInstrLevel(Module module){
        for(var function: module.functions.values()){
            for(var basicBlock:function.blocks){
                getUseFulInstr(basicBlock);
            }
        }
        for(var function: module.functions.values()){
            for(var basicBlock:function.blocks){
                Iterator<Value> it = basicBlock.getInsts().iterator();
                while (it.hasNext()) {
                    Instr instr = (Instr) it.next();
                    if (!usefulInstrClosure.contains(instr))
                    {
                        instr.delete();
                        it.remove();
                    }
                }
            }
        }
    }

    public boolean isUseFulInstr(Value instr){
        return  instr instanceof Store ||
                instr instanceof Br    ||
                instr instanceof Ret   ||
                (instr instanceof Call && ((Call) instr).getFunction().hasSideEffect);
    }
    public void getUseFulInstrClosure(Value instr){
        if(usefulInstrClosure.contains(instr)) return;
        usefulInstrClosure.add(instr);
        for(var operand:((Instr) instr).getOperands()){
            if(operand instanceof Instr) queue.add((Instr)operand);
        }
    }
    public void getUseFulInstr(BasicBlock basicBlock){
        for(var instr:basicBlock.getInsts()){
            if(isUseFulInstr(instr)) {
                queue.addLast((User)instr);
            }

        }
        while(!queue.isEmpty()){
            getUseFulInstrClosure(queue.pop());
        }
    }

    public void deadCodeEmitFunctionLevel(Module module){
        var main = module.getFunction("main");
        getUsefulFunctionClosure(main);
//        for(var function:usefulFunctionClosure) System.out.println(function.getFullName());
        for(var key: module.functions.keySet()){
            if(!usefulFunctionClosure.contains(module.functions.get(key))&&module.functions.get(key).blocks.size()!=0){
                uselessFunction.add(module.functions.get(key));
            }
        }
        for(var function: uselessFunction){
            module.functions.remove(function.name);
        }
    }

    public void getUsefulFunctionClosure(Function function) {
        if(usefulFunctionClosure.contains(function)) return;
        usefulFunctionClosure.add(function);
        for(var callee:function.getCallees()){
            getUsefulFunctionClosure(callee);
        }
    }


}

